@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 1999, 2010 Tanuki Software, Ltd.
+ * Copyright (c) 1999, 2012 Tanuki Software, Ltd.
  * http://www.tanukisoftware.com
  * All rights reserved.
  *
@@ -78,6 +78,10 @@
 #define WRAPPER_TIMER_FAST_THRESHOLD (2 * 24 * 3600 * 1000 / WRAPPER_TICK_MS) /* Default to 2 days. */
 #define WRAPPER_TIMER_SLOW_THRESHOLD (2 * 24 * 3600 * 1000 / WRAPPER_TICK_MS) /* Default to 2 days. */
 
+#define WRAPPER_BACKEND_TYPE_UNKNOWN 0 /* Unknown type. */
+#define WRAPPER_BACKEND_TYPE_SOCKET  1 /* Use a loopback socket to communicate. */
+#define WRAPPER_BACKEND_TYPE_PIPE    2 /* Use a pair of pipes to communicate. */
+
 #define WRAPPER_WSTATE_STARTING  51 /* Wrapper is starting.  Remains in this state
                                      *  until the JVM enters the STARTED state or
                                      *  the wrapper jumps into the STOPPING state
@@ -101,37 +105,39 @@
 #define WRAPPER_JSTATE_DOWN_CHECK 70 /* JVM is confirmed to be down, but we still need
                                      *  to do our cleanup work.  This is the state after
                                      *  a JVM process has gone away. */
-#define WRAPPER_JSTATE_DOWN_CLEAN 71 /* JVM is confirmed to be down and we have cleaned
-                                     *  up.  This is the initial state and the state
-                                     *  after the JVM process has gone away and and
-                                     *  cleanup has been done. */
-#define WRAPPER_JSTATE_LAUNCH_DELAY 72 /* Set from the DOWN state to launch a JVM.  The
+#define WRAPPER_JSTATE_DOWN_FLUSH 71 /* JVM is confirmed to be down but we still need
+                                     *  flush and process all of its output. */
+#define WRAPPER_JSTATE_DOWN_CLEAN 72 /* JVM is confirmed to be down and we have cleaned
+                                     *  up and flushed all output.  This is the initial
+                                     *  state and the state after the JVM process has
+                                     *  gone away and and cleanup has been done. */
+#define WRAPPER_JSTATE_LAUNCH_DELAY 73 /* Set from the DOWN state to launch a JVM.  The
                                      *  timeout will be the time to actually launch
                                      *  the JVM after any required delay. */
-#define WRAPPER_JSTATE_RESTART   73 /* JVM is about to be restarted. No timeout. */
-#define WRAPPER_JSTATE_LAUNCH    74 /* JVM is about to launch a JVM. No timeout. */
-#define WRAPPER_JSTATE_LAUNCHING 75 /* JVM was launched, but has not yet responded.
+#define WRAPPER_JSTATE_RESTART   74 /* JVM is about to be restarted. No timeout. */
+#define WRAPPER_JSTATE_LAUNCH    75 /* JVM is about to launch a JVM. No timeout. */
+#define WRAPPER_JSTATE_LAUNCHING 76 /* JVM was launched, but has not yet responded.
                                      *  Must enter the LAUNCHED state before <t>
                                      *  or the JVM will be killed. */
-#define WRAPPER_JSTATE_LAUNCHED  76 /* JVM was launched, and responed to a ping. */
-#define WRAPPER_JSTATE_STARTING  77 /* JVM has been asked to start.  Must enter the
+#define WRAPPER_JSTATE_LAUNCHED  77 /* JVM was launched, and responed to a ping. */
+#define WRAPPER_JSTATE_STARTING  78 /* JVM has been asked to start.  Must enter the
                                      *  STARTED state before <t> or the JVM will be
                                      *  killed. */
-#define WRAPPER_JSTATE_STARTED   78 /* JVM has responded that it is running.  Must
+#define WRAPPER_JSTATE_STARTED   79 /* JVM has responded that it is running.  Must
                                      *  respond to a ping by <t> or the JVM will
                                      *  be killed. */
-#define WRAPPER_JSTATE_STOP      79 /* JVM is about to be sent a stop command to shutdown
+#define WRAPPER_JSTATE_STOP      80 /* JVM is about to be sent a stop command to shutdown
                                      *  cleanly. */
-#define WRAPPER_JSTATE_STOPPING  80 /* JVM was sent a stop command, but has not yet
+#define WRAPPER_JSTATE_STOPPING  81 /* JVM was sent a stop command, but has not yet
                                      *  responded.  Must enter the STOPPED state
                                      *  and exit before <t> or the JVM will be killed. */
-#define WRAPPER_JSTATE_STOPPED   81 /* JVM has responed that it is stopped. */
-#define WRAPPER_JSTATE_KILLING   82 /* The Wrapper is about ready to kill the JVM
+#define WRAPPER_JSTATE_STOPPED   82 /* JVM has responed that it is stopped. */
+#define WRAPPER_JSTATE_KILLING   83 /* The Wrapper is about ready to kill the JVM
                                      *  process but it must wait a few moments before
                                      *  actually doing so.  After <t> has expired, the
                                      *  JVM will be killed and we will enter the STOPPED
                                      *  state. */
-#define WRAPPER_JSTATE_KILL      83 /* The Wrapper is about ready to kill the JVM process. */
+#define WRAPPER_JSTATE_KILL      84 /* The Wrapper is about ready to kill the JVM process. */
 
 /* Defined Action types.  Registered actions are negative.  Custom types are positive. */
 #define ACTION_LIST_END          0
@@ -142,6 +148,8 @@
 #define ACTION_DEBUG             -5
 #define ACTION_PAUSE             -6
 #define ACTION_RESUME            -7
+#define ACTION_SUCCESS           -8
+#define ACTION_GC                -9
 #if defined(MACOSX)
 #define TRIGGER_ADVICE_NIL_SERVER TEXT("****** Returning nil _server **********")
 #define ACTION_ADVICE_NIL_SERVER -32
@@ -179,19 +187,28 @@ typedef unsigned long TICKS;
 #define WRAPPER_ENV_SOURCE_REG_ACCOUNT 8
 #endif
 
+#ifdef WIN32
+/* Defines the maximum number of service manager control events that can be queued in a single loop. */
+#define CTRL_CODE_QUEUE_SIZE 26 /* Can enqueue one less than this count at any time. */
+#endif
 
 /* Type definitions */
 typedef struct WrapperConfig WrapperConfig;
 struct WrapperConfig {
-    TCHAR*   argCommand;             /* The command used to launch the wrapper. */
-    TCHAR*   argCommandArg;          /* The argument to the command used to launch the wrapper. */
-    TCHAR*   argConfFile;            /* The name of the config file from the command line. */
+    TCHAR   *argCommand;            /* The command used to launch the wrapper. */
+    TCHAR   *argCommandArg;         /* The argument to the command used to launch the wrapper. */
+    TCHAR   *argConfFile;           /* The name of the config file from the command line. */
+    TCHAR   *confDir;
     int     argConfFileDefault;     /* True if the config file was not specified. */
     int     argConfFileFound;       /* True if the config file was found. */
     int     argCount;               /* The total argument count. */
-    TCHAR**  argValues;              /* Argument values. */
+    TCHAR   **argValues;            /* Argument values. */
+    TCHAR   **javaArgValues;        /* Arguments getting passed over to the java application */
+    int     javaArgValueCount;      /* Number of the arguments getting passed over to the java application */
 
-    TCHAR*   language;               /* The language */
+    TCHAR   *initialPath;           /* What the working directory was when the Wrapper process was first launched. */
+    TCHAR   *language;              /* The language */
+    int     backendType;            /* The type of the backend that the Wrapper and Java use to communicate. */
     int     configured;             /* TRUE if loadConfiguration has been called. */
     int     useSystemTime;          /* TRUE if the wrapper should use the system clock for timing, FALSE if a tick counter should be used. */
     int     timerFastThreshold;     /* If the difference between the system time based tick count and the timer tick count ever falls by more than this value then a warning will be displayed. */
@@ -208,20 +225,21 @@ struct WrapperConfig {
     int     jvmPortMin;             /* Minimum port which the JVM should bind to when connecting back to the wrapper. */
     int     jvmPortMax;             /* Maximum port which the JVM should bind to when connecting back to the wrapper. */
     int     sock;                   /* Socket number. if open. */
-    TCHAR    *originalWorkingDir;    /* Original Wrapper working directory. */
-    TCHAR    *workingDir;            /* Configured working directory. */
-    TCHAR    *configFile;            /* Name of the configuration file */
+    TCHAR   *originalWorkingDir;    /* Original Wrapper working directory. */
+    TCHAR   *workingDir;            /* Configured working directory. */
+    TCHAR   *configFile;            /* Name of the configuration file */
     int     commandLogLevel;        /* The log level to use when logging the java command. */
 #ifdef WIN32
-    TCHAR    *jvmCommand;            /* Command used to launch the JVM */
+    TCHAR   *jvmCommand;            /* Command used to launch the JVM */
 #else /* UNIX */
-    TCHAR    **jvmCommand;           /* Command used to launch the JVM */
+    TCHAR   **jvmCommand;           /* Command used to launch the JVM */
 #endif
+    int     detachStarted;          /* TRUE if the JVM process should be detached once it has started. */
     int     environmentClasspath;   /* TRUE if the classpath should be passed to the JVM in the environment. */
-    TCHAR    *classpath;             /* Classpath to pass to the JVM. */
+    TCHAR   *classpath;             /* Classpath to pass to the JVM. */
     int     debugJVM;               /* True if the JVM is being launched with a debugger enabled. */
     int     debugJVMTimeoutNotified;/* True if the JVM is being launched with a debugger enabled and the user has already been notified of a timeout. */
-    TCHAR    key[17];                /* Key which the JVM uses to authorize connections. (16 digits + \0) */
+    TCHAR   key[17];                /* Key which the JVM uses to authorize connections. (16 digits + \0) */
     int     isConsole;              /* TRUE if the wrapper was launched as a console. */
     int     cpuTimeout;             /* Number of seconds without CPU before the JVM will issue a warning and extend timeouts */
     int     startupTimeout;         /* Number of seconds the wrapper will wait for a JVM to startup */
@@ -231,10 +249,15 @@ struct WrapperConfig {
     int     shutdownTimeout;        /* Number of seconds the wrapper will wait for a JVM to shutdown */
     int     jvmExitTimeout;         /* Number of seconds the wrapper will wait for a JVM to process to terminate */
     int     jvmCleanupTimeout;      /* Number of seconds the wrapper will allow for its post JVM shudown cleanup. */
-    int     isJVMOrphaned;          /* If set to true, the Wrapper goes into a test more where it stops communicating with the JVM. */
+    int     useJavaIOThread;        /* If TRUE then a dedicated thread will be used to process console output form the JVM. */
+    int     pauseThreadMain;        /* Number of seconds to pause the main thread on its next loop.  Only used for testing. */
+    int     pauseThreadTimer;       /* Number of seconds to pause the timer thread on its next loop.  Only used for testing. */
+    int     pauseThreadJavaIO;      /* Number of seconds to pause the javaio thread on its next loop.  Only used for testing. */
 
 #ifdef WIN32
     int     ignoreUserLogoffs;      /* If TRUE, the Wrapper will ignore logoff events when run in the background as an in console mode. */
+    TCHAR   *userName;              /* The username (account) of the Wrapper process. */
+    TCHAR   *domainName;            /* The domain of the Wrapper process. */
     DWORD   wrapperPID;             /* PID of the Wrapper process. */
     DWORD   javaPID;                /* PID of the Java process. */
     HANDLE  wrapperProcess;         /* Handle of the Wrapper process. */
@@ -252,9 +275,10 @@ struct WrapperConfig {
 
     int     isDebugging;            /* TRUE if set in the configuration file */
     int     isAdviserEnabled;       /* TRUE if advice messages should be output. */
-    const TCHAR *nativeLibrary;      /* The base name of the native library loaded by the WrapperManager. */
+    const TCHAR *nativeLibrary;     /* The base name of the native library loaded by the WrapperManager. */
     int     libraryPathAppendPath;  /* TRUE if the PATH environment variable should be appended to the java library path. */
     int     isStateOutputEnabled;   /* TRUE if set in the configuration file.  Shows output on the state of the state engine. */
+    int     isJavaIOOutputEnabled;  /* TRUE if detailed javaIO output should be included in debug output. */
     int     isTickOutputEnabled;    /* TRUE if detailed tick timer output should be included in debug output. */
     int     isLoopOutputEnabled;    /* TRUE if very detailed output from the main loop should be output. */
     int     isSleepOutputEnabled;   /* TRUE if detailed sleep output should be included in debug output. */
@@ -264,39 +288,48 @@ struct WrapperConfig {
     int     isCPUOutputEnabled;     /* TRUE if detailed CPU output should be included in status output. */
     int     cpuOutputInterval;      /* Interval in seconds at which CPU usage is logged. */
     TICKS   cpuOutputTimeoutTicks;  /* Tick count at which CPU will next be logged. */
+    int     isPageFaultOutputEnabled;/* TRUE if detailed PageFault output should be included in status output. */
+    int     pageFaultOutputInterval;/* Interval in seconds at which PageFault usage is logged. */
+    TICKS   pageFaultOutputTimeoutTicks; /* Tick count at which PageFault will next be logged. */
     int     logfileInactivityTimeout; /* The number of seconds of inactivity before the logfile will be closed. */
     TICKS   logfileInactivityTimeoutTicks; /* Tick count at which the logfile will be considered inactive and closed. */
-    int     isShutdownHookDisabled; /* TRUE if set in the configuration file */
+    int     isShutdownHookDisabled; /* TRUE if the use of a shutdown hook by the WrapperManager class should be disabled. */
+    int     isForcedShutdownDisabled; /* TRUE if forced shutdowns are disabled. */
     int     startupDelayConsole;    /* Delay in seconds before starting the first JVM in console mode. */
     int     startupDelayService;    /* Delay in seconds before starting the first JVM in service mode. */
     int     exitCode;               /* Code which the wrapper will exit with */
     int     exitRequested;          /* TRUE if the current JVM should be shutdown. */
     int     restartRequested;       /* WRAPPER_RESTART_REQUESTED_NO, WRAPPER_RESTART_REQUESTED_AUTOMATIC, or WRAPPER_RESTART_REQUESTED_CONFIGURED if the another JVM should be launched after the current JVM is shutdown. Only set if exitRequested is set. */
+    int     stoppedPacketReceived;  /* TRUE if the STOPPED packet was received before a restart. */
+    int     restartPacketReceived;  /* TRUE if the RESTART packet was received before a restart. */
     int     jvmRestarts;            /* Number of times that a JVM has been launched since the wrapper was started. */
     int     restartDelay;           /* Delay in seconds before restarting a new JVM. */
     int     restartReloadConf;      /* TRUE if the configuration should be reloaded before a JVM restart. */
     int     isRestartDisabled;      /* TRUE if restarts should be disabled. */
     int     isAutoRestartDisabled;  /* TRUE if automatic restarts should be disabled. */
     int     requestThreadDumpOnFailedJVMExit; /* TRUE if the JVM should be asked to dump its state when it fails to halt on request. */
+    int     requestThreadDumpOnFailedJVMExitDelay; /* Number of seconds to wait after the thread dump before killing the JVM. */
     TICKS   jvmLaunchTicks;         /* The tick count at which the previous or current JVM was launched. */
     int     failedInvocationCount;  /* The number of times that the JVM exited in less than successfulInvocationTime in a row. */
     int     successfulInvocationTime;/* Amount of time that a new JVM must be running so that the invocation will be considered to have been a success, leading to a reset of the restart count. */
     int     maxFailedInvocations;   /* Maximum number of failed invocations in a row before the Wrapper will give up and exit. */
     int     outputFilterCount;      /* Number of registered output filters. */
-    TCHAR**  outputFilters;          /* Array of output filters. */
-    int**   outputFilterActionLists;/* Array of output filter action lists. */
-    TCHAR**  outputFilterMessages;   /* Array of output filter messages. */
-    TCHAR    *pidFilename;           /* Name of file to store wrapper pid in */
-    TCHAR    *lockFilename;          /* Name of file to store wrapper lock in */
-    TCHAR    *javaPidFilename;       /* Name of file to store jvm pid in */
-    TCHAR    *javaIdFilename;        /* Name of file to store jvm id in */
-    TCHAR    *statusFilename;        /* Name of file to store wrapper status in */
-    TCHAR    *javaStatusFilename;    /* Name of file to store jvm status in */
-    TCHAR    *commandFilename;       /* Name of a command file used to send commands to the Wrapper. */
+    TCHAR   **outputFilters;        /* Array of output filters. */
+    int     **outputFilterActionLists;/* Array of output filter action lists. */
+    TCHAR   **outputFilterMessages; /* Array of output filter messages. */
+    int     *outputFilterAllowWildFlags; /* Array of output filter flags that say whether or not wild cards in the filter can be processed. */
+    size_t  *outputFilterMinLens;   /* Array of the minimum text lengths that could possibly match the specified filter.  Only used if it contains wildcards. */
+    TCHAR   *pidFilename;           /* Name of file to store wrapper pid in */
+    TCHAR   *lockFilename;          /* Name of file to store wrapper lock in */
+    TCHAR   *javaPidFilename;       /* Name of file to store jvm pid in */
+    TCHAR   *javaIdFilename;        /* Name of file to store jvm id in */
+    TCHAR   *statusFilename;        /* Name of file to store wrapper status in */
+    TCHAR   *javaStatusFilename;    /* Name of file to store jvm status in */
+    TCHAR   *commandFilename;       /* Name of a command file used to send commands to the Wrapper. */
     int     commandFileTests;       /* True if test commands will be accepted via the command file. */
     int     commandPollInterval;    /* Interval in seconds at which the existence of the command file is polled. */
     TICKS   commandTimeoutTicks;    /* Tick count at which the command file will be checked next. */
-    TCHAR    *anchorFilename;        /* Name of an anchor file used to control when the Wrapper should quit. */
+    TCHAR   *anchorFilename;        /* Name of an anchor file used to control when the Wrapper should quit. */
     int     anchorPollInterval;     /* Interval in seconds at which the existence of the anchor file is polled. */
     TICKS   anchorTimeoutTicks;     /* Tick count at which the anchor file will be checked next. */
     int     umask;                  /* Default umask for all files. */
@@ -309,24 +342,27 @@ struct WrapperConfig {
     int     javaStatusFileUmask;    /* Umask to use when creating the java status file. */
     int     anchorFileUmask;        /* Umask to use when creating the anchor file. */
     int     ignoreSignals;          /* Mask that determines where the Wrapper should ignore any catchable system signals.  Can be ingored in the Wrapper and/or JVM. */
-    TCHAR    *consoleTitle;          /* Text to set the console title to. */
-    TCHAR    *serviceName;           /* Name of the service. */
-    TCHAR    *serviceDisplayName;    /* Display name of the service. */
-    TCHAR    *serviceDescription;    /* Description for service. */
-    TCHAR    *hostName;              /* The name of the current host. */
+    TCHAR   *consoleTitle;          /* Text to set the console title to. */
+    TCHAR   *serviceName;           /* Name of the service. */
+    TCHAR   *serviceDisplayName;    /* Display name of the service. */
+    TCHAR   *serviceDescription;    /* Description for service. */
+    TCHAR   *hostName;              /* The name of the current host. */
     int     pausable;               /* Should the service be allowed to be paused? */
     int     pausableStopJVM;        /* Should the JVM be stopped when the service is paused? */
+    int     initiallyPaused;        /* Should the Wrapper come up initially in a paused state? */
+    int     logLFDelayThreshold;    /* The LF Delay threshold to use when logging java output. */
 
 #ifdef WIN32
     int     isSingleInvocation;     /* TRUE if only a single invocation of an application should be allowed to launch. */
-    TCHAR    *ntServiceLoadOrderGroup; /* Load order group name. */
-    TCHAR    *ntServiceDependencies; /* List of Dependencies */
+    TCHAR   *ntServiceLoadOrderGroup; /* Load order group name. */
+    TCHAR   *ntServiceDependencies; /* List of Dependencies */
     int     ntServiceStartType;     /* Mode in which the Service is installed.
                                      * {SERVICE_AUTO_START | SERVICE_DEMAND_START} */
     DWORD   ntServicePriorityClass; /* Priority at which the Wrapper and its JVMS will run.
                                      * {HIGH_PRIORITY_CLASS | IDLE_PRIORITY_CLASS | NORMAL_PRIORITY_CLASS | REALTIME_PRIORITY_CLASS} */
-    TCHAR    *ntServiceAccount;      /* Account name to use when running as a service.  NULL to use the LocalSystem account. */
-    TCHAR    *ntServicePassword;     /* Password to use when running as a service.  NULL means no password. */
+    TCHAR   *ntServiceAccount;      /* Account name to use when running as a service.  NULL to use the LocalSystem account. */
+    TCHAR   *ntServicePassword;     /* Password to use when running as a service.  NULL means no password. */
+    int     ntServicePrompt;        /* If true then the user will be prompted for a account name, domain,  password when installing as a service. */
     int     ntServicePasswordPrompt; /* If true then the user will be prompted for a password when installing as a service. */
     int     ntServicePasswordPromptMask; /* If true then the password will be masked as it is input. */
     int     ntServiceInteractive;   /* Should the service be allowed to interact with the desktop? */
@@ -353,7 +389,10 @@ struct WrapperConfig {
     int     ctrlEventCloseTrapped;  /* CTRL_CLOSE_EVENT trapped. */
     int     ctrlEventLogoffTrapped; /* CTRL_LOGOFF_EVENT trapped. */
     int     ctrlEventShutdownTrapped;/* CTRL_SHUTDOWN_EVENT trapped. */
-    int     ctrlCodeLast;           /* Id of the last control code trapped. */
+    int     *ctrlCodeQueue;         /* Queue of control code ids trapped. */
+    int     ctrlCodeQueueWriteIndex;
+    int     ctrlCodeQueueReadIndex;
+    int     ctrlCodeQueueWrapped;
     int     ctrlCodePauseTrapped;   /* SERVICE_CONTROL_PAUSE was trapped. */
     int     ctrlCodeContinueTrapped;/* SERVICE_CONTROL_CONTINUE was trapped. */
     int     ctrlCodeStopTrapped;    /* SERVICE_CONTROL_STOP was trapped. */
@@ -386,16 +425,16 @@ struct WrapperConfig {
 #define WRAPPER_MSG_KEY           (char)110
 #define WRAPPER_MSG_BADKEY        (char)111
 #define WRAPPER_MSG_LOW_LOG_LEVEL (char)112
-#define WRAPPER_MSG_PING_TIMEOUT  (char)113
+#define WRAPPER_MSG_PING_TIMEOUT  (char)113 /* No longer used. */
 #define WRAPPER_MSG_SERVICE_CONTROL_CODE (char)114
 #define WRAPPER_MSG_PROPERTIES    (char)115
 /** Log commands are actually 116 + the LOG LEVEL (LEVEL_UNKNOWN ~ LEVEL_NONE), (116 ~ 124). */
 #define WRAPPER_MSG_LOG           (char)116
 #define WRAPPER_MSG_LOGFILE       (char)134
-#define WRAPPER_MSG_APPEAR_ORPHAN (char)137
+#define WRAPPER_MSG_APPEAR_ORPHAN (char)137 /* No longer used. */
 #define WRAPPER_MSG_PAUSE         (char)138
 #define WRAPPER_MSG_RESUME        (char)139
-
+#define WRAPPER_MSG_GC            (char)140
 #define WRAPPER_PROCESS_DOWN      200
 #define WRAPPER_PROCESS_UP        201
 
@@ -421,17 +460,17 @@ extern void wrapperProtocolClose();
 extern int wrapperProtocolFunction(char function, const TCHAR *message);
 
 /**
- * Checks the status of the server socket.
+ * Checks the status of the server backend.
  *
- * The socket will be initialized if the JVM is in a state where it should
- *  be up, otherwise the socket will be left alone.
+ * The backend will be initialized if the JVM is in a state where it should
+ *  be up, otherwise the backend will be left alone.
  *
  * If the forceOpen flag is set then an attempt will be made to initialize
- *  the socket regardless of the JVM state.
+ *  the backend regardless of the JVM state.
  *
- * Returns TRUE if the socket is open and ready on return, FALSE if not.
+ * Returns TRUE if the backend is open and ready on return, FALSE if not.
  */
-extern int wrapperCheckServerSocket(int forceOpen);
+extern int wrapperCheckServerBackend(int forceOpen);
 
 /**
  * Read any data sent from the JVM.  This function will loop and read as many
@@ -445,9 +484,69 @@ extern int wrapperProtocolRead();
 /******************************************************************************
  * Utility Functions
  *****************************************************************************/
+/**
+ * Test function to pause the current thread for the specified amount of time.
+ *  This is used to test how the rest of the Wrapper behaves when a particular
+ *  thread blocks for any reason.
+ *
+ * @param pauseTime Number of seconds to pause for.  -1 will pause indefinitely.
+ * @param threadName Name of the thread that will be logged prior to pausing.
+ */
+extern void wrapperPauseThread(int pauseTime, const TCHAR *threadName);
+
+/**
+ * Function that will recursively attempt to match two strings where the
+ *  pattern can contain '?' or '*' wildcard characters.
+ *
+ * @param text Text to be searched.
+ * @param pattern Pattern to search for.
+ * @param patternLen Length of the pattern.
+ * @param minTextLen Minimum number of characters that the text needs to possibly match the pattern.
+ *
+ * @return TRUE if found, FALSE otherwise.
+ */
+extern int wrapperWildcardMatch(const TCHAR *text, const TCHAR *pattern, size_t minTextLen);
+
+/**
+ * Calculates the minimum text length which could be matched by the specified pattern.
+ *  Patterns can contain '*' or '?' wildcards.
+ *  '*' matches 0 or more characters.
+ *  '?' matches exactly one character.
+ *
+ * @param pattern Pattern to calculate.
+ *
+ * @return The minimum text length of the pattern.
+ */
+extern size_t wrapperGetMinimumTextLengthForPattern(const TCHAR *pattern);
+
+/**
+ * Returns a constant text representation of the specified Wrapper State.
+ *
+ * @param wState The Wrapper State whose name is being requested.
+ *
+ * @return Thre requested Wrapper State.
+ */
+extern const TCHAR *wrapperGetWState(int wState);
+
+/**
+ * Returns a constant text representation of the specified Java State.
+ *
+ * @param jState The Java State whose name is being requested.
+ *
+ * @return Thre requested Java State.
+ */
+extern const TCHAR *wrapperGetJState(int jState);
+
 extern struct tm wrapperGetReleaseTime();
 extern struct tm wrapperGetBuildTime();
 
+#ifdef WIN32
+extern int initializeStartup();
+extern void disposeStartup();
+#endif
+extern void disposeJavaIO();
+extern void disposeTimer();
+
 extern int showHostIds(int logLevel);
 extern void wrapperLoadHostName();
 
@@ -477,11 +576,17 @@ extern void wrapperProcessActionList(int *actionList, const TCHAR *triggerMsg, i
 
 extern void wrapperAddDefaultProperties();
 
-extern int wrapperLoadConfigurationProperties();
+extern int wrapperLoadConfigurationProperties(int preload);
 
 extern void wrapperGetCurrentTime(struct timeb *timeBuffer);
 
 #ifdef WIN32
+
+extern void wrapperInitializeProfileCounters();
+extern void wrapperDumpPageFaultUsage();
+
+extern void updateStringValue(TCHAR **ptr, const TCHAR *value);
+
 extern TCHAR** wrapperGetSystemPath();
 extern int wrapperGetJavaHomeFromWindowsRegistry(TCHAR *javaHome);
 #endif
@@ -577,10 +682,16 @@ extern void wrapperSetJavaState(int jState, TICKS nowTicks, int delay);
  *****************************************************************************/
 #ifdef WIN32
 extern void wrapperCheckConsoleWindows();
+/**
+ *   checks the digital Signature of the binary and reports the result.
+ */
+extern BOOL verifyEmbeddedSignature();
 
 extern int exceptionFilterFunction(PEXCEPTION_POINTERS exceptionPointers);
-BOOL extern myShellExec(HWND hwnd, LPCTSTR pszVerb, LPCTSTR pszPath, LPCTSTR pszParameters, LPCTSTR pszDirectory);
-BOOL extern runElevated( __in LPCTSTR pszPath, __in_opt LPCTSTR pszParameters, __in_opt LPCTSTR pszDirectory);
+BOOL extern elevateThis(int argc, TCHAR **argv);
+BOOL extern duplicateSTD();
+BOOL extern myShellExec(HWND hwnd, LPCTSTR pszVerb, LPCTSTR pszPath, LPCTSTR pszParameters, LPCTSTR pszDirectory, TCHAR* namedPipeName);
+BOOL extern runElevated( __in LPCTSTR pszPath, __in_opt LPCTSTR pszParameters, __in_opt LPCTSTR pszDirectory, TCHAR* namedPipeName);
 BOOL extern isElevated();
 BOOL extern isVista();
 extern void wrapperMaintainControlCodes();
@@ -601,8 +712,12 @@ extern int wrapperInitializeRun();
 /**
  * Cause the current thread to sleep for the specified number of milliseconds.
  *  Sleeps over one second are not allowed.
+ *
+ * @param ms Number of milliseconds to wait for.
+ *
+ * @return TRUE if the was interrupted, FALSE otherwise.  Neither is an error.
  */
-extern void wrapperSleep(int useLoggerQueue, int ms);
+extern int wrapperSleep(int ms);
 
 /**
  * Reports the status of the wrapper to the service manager
@@ -712,15 +827,31 @@ extern void wrapperPauseProcess(int actionCode);
 extern void wrapperResumeProcess(int actionCode);
 
 /**
- * Used to ask the state engine to shut down the JVM and Wrapper
+ * Detaches the Java process so the Wrapper will if effect forget about it.
  */
-extern void wrapperStopProcess(int exitCode);
+extern void wrapperDetachJava();
+
+/**
+ * Used to ask the state engine to shut down the JVM and Wrapper.
+ *
+ * @param exitCode Exit code to use when shutting down.
+ * @param force True to force the Wrapper to shutdown even if some configuration
+ *              had previously asked that the JVM be restarted.  This will reset
+ *              any existing restart requests, but it will still be possible for
+ *              later actions to request a restart.
+ */
+extern void wrapperStopProcess(int exitCode, int force);
 
 /**
  * Used to ask the state engine to shut down the JVM.
  */
 extern void wrapperRestartProcess();
 
+/**
+ * Sends a command off to the JVM asking it to perform a garbage collection sweep.
+ */
+extern void wrapperRequestJVMGC();
+
 /**
  * Loops over and strips all double quotes from prop and places the
  *  stripped version into propStripped.
